/*
 *   MIRACL utility program to automatically generate a mirdef.h file 
 *
 *   Run this with your computer/compiler configuration. It will
 *   attempt to create best-fit mirdef.h file for compiling the MIRACL 
 *   library. 
 *
 *   Copyright (c) 1994-1999 Shamus Software Ltd.
 */

#include <stdio.h>
#include <stdlib.h>

static int answer(void)
{
    char ch;
    scanf("%c",&ch);
    getchar();
    if (ch=='Y' || ch=='y') return 1;
    return 0;
}

void main()
{
    int chosen,utlen,dllen,flsh,stripped,standard,nofull,port,mant,r;
    int iobsiz,nbits,words,i,b,dlong;
    int end;
    char *ptr;
    char longlong[20];
    long s;
    double x,y,eps=1.0;
    FILE *fp;
    printf("This program attempts to generate a mirdef.h file from your\n");
    printf("responses to some simple questions, and by some internal tests\n");
    printf("of its own.\n\n");
    printf("Read mrmuldv.any CAREFULLY, for useful insights and advice.\n\n");
    printf("The most fundamental decision is that of the 'underlying type'\n");
    printf("that is the C data type to be used to store each digit of a\n");
    printf("big number. You input the number of bits in this type, and \n");
    printf("this program finds a suitable candidate (usually short, int or \n");
    printf("long). Typical values would be 16, 32 or perhaps 64. \n");
    printf("The bigger the better, but a good starting point would be to\n");
    printf("enter the native wordlength of your computer\n\n");
    printf("Now enter number of bits in underlying type= ");
    scanf("%d",&utlen);
    getchar();
    printf("\n");
    if (utlen!=8 && utlen!=16 && utlen!=32 && utlen!=64) 
    {
        printf("Non-standard system - this program cannot help!\n");
        exit(0);
    }
    fp=fopen("mirdef.tst","w");
    fprintf(fp,"/*\n *   MIRACL compiler/hardware definitions - mirdef.h\n");
    fprintf(fp," *   Copyright (c) 1988-1997 Shamus Software Ltd.\n */\n\n");
    fprintf(fp,"\n#define MIRACL %d\n",utlen);


    end=1;
    ptr=(char *)(&end);

    if ((int)(*ptr) == 1)
    {
         printf("    Little-endian processor detected\n\n");
         fprintf(fp,"#define MR_LITTLE_ENDIAN\n");
    }
    else 
    {
         printf("    Big-endian processor detected\n\n");
         fprintf(fp,"#define MR_BIG_ENDIAN\n");

    }


    chosen=0;
    if (utlen==8)
    {
        printf("    underlying type is a char\n");
        fprintf(fp,"#define mr_utype char\n");
        chosen=1;
    }
    if (!chosen && utlen==8*sizeof(short))
    {
        printf("    underlying type is a short\n");
        fprintf(fp,"#define mr_utype short\n");
        chosen=1;
    }
    if (!chosen && utlen==8*sizeof(int))
    {
        printf("    underlying type is an int\n");
        fprintf(fp,"#define mr_utype int\n");
        chosen=1;
    }
    if (!chosen && utlen==8*sizeof(long))
    {
        printf("    underlying type is a long\n");
        fprintf(fp,"#define mr_utype long\n");
        chosen=1;
    }
    dlong=0;
    if (!chosen && utlen==16*sizeof(long))
    {
        printf("\nDoes compiler support a %d bit unsigned type? (Y/N)?", 
              utlen);
        dlong=answer();
        if (dlong)
        {
            printf("Is it called long long? (Y/N)?");
            if (!answer())
            {
                printf("Enter its name : ");
                scanf("%s",longlong);
                getchar();
            }
            else strcpy(longlong,"long long");
            printf("    underlying type is a %s\n",longlong);
            fprintf(fp,"#define mr_utype %s\n",longlong);
            chosen=1;
        }
    }

    if (!chosen)
    {
        printf("No suitable underlying type could be found\n");
        fclose(fp);
        exit(0);
    }

/* Now try to find a 32-bit unsigned type */

 
    chosen=0;
    if (32==8*sizeof(unsigned short))
    {
        printf("    32 bit unsigned type is a unsigned short\n");
        fprintf(fp,"#define mr_unsign32 unsigned short\n");
        chosen=1;
    }
    if (!chosen && 32==8*sizeof(unsigned int))
    {
        printf("    32 bit unsigned type is an unsigned int\n");
        fprintf(fp,"#define mr_unsign32 unsigned int\n");
        chosen=1;
    }
    if (!chosen && 32==8*sizeof(long))
    {
        printf("    32 bit unsigned type is an unsigned long\n");
        fprintf(fp,"#define mr_unsign32 unsigned long\n");
        chosen=1;
    }

    if (!chosen)
    {
        printf("No suitable 32 bit unsigned type could be found\n");
        fclose(fp);
        exit(0);
    }

/* Now try to find a double length type */

    dllen=2*utlen;
    chosen=0;
    if (!chosen && dllen==8*sizeof(short))
    {
        printf("    double length type is a short\n");
        fprintf(fp,"#define mr_dltype short\n");
        chosen=1;
    }
    if (!chosen && dllen==8*sizeof(int))
    {
        printf("    double length type is an int\n");
        fprintf(fp,"#define mr_dltype int\n");
        chosen=1;
    }
    if (!chosen && dllen==8*sizeof(long))
    {
        printf("    double length type is a long\n");
        fprintf(fp,"#define mr_dltype long\n");
        chosen=1;
    }
    dlong=0;
    if (!chosen && dllen==16*sizeof(long))
    {
        printf("\nDoes compiler support a %d bit type? (Y/N)?", 
              dllen);
        dlong=answer();
        if (dlong)
        {
            printf("Is it called long long? (Y/N)?");
            if (!answer())
            {
                printf("Enter its name : ");
                scanf("%s",longlong);
                getchar();
            }
            else strcpy(longlong,"long long");

            printf("    double length type is a %s\n",longlong);
            fprintf(fp,"#define mr_dltype %s\n",longlong);
            chosen=1;
        }
    }
    port=0;
    if (chosen)
    {
        printf("\nDo you want a C-only version of MIRACL (Y/N)?");
        port=answer();
        if (port) fprintf(fp,"#define MR_NOASM\n");
    }

    printf("Do you want support for flash arithmetic? (Y/N)?");
    flsh=answer();
    if (flsh)
    { /* calculate size of mantissa in bits */
        for (mant=0;;mant++)
        { /* IMPORTANT TO FOOL OPTIMIZER!!!!!! */
            x=1.0+eps;
            y=1.0;
            if (x==y) break;
            eps/=2.0;
        }
        mant--;
        fprintf(fp,"#define MR_FLASH %d\n",mant);  
    }

    printf("Do you want stripped-down version (smaller - no error messages) (Y/N)?");
    stripped=answer();
    if (stripped) fprintf(fp,"#define MR_STRIPPED_DOWN\n");

    printf("Does your development environment support standard screen/keyboard I/O?\n");
    printf("(It doesn't for example in MS Windows, and embedded applications)\n");
    printf("If in doubt, answer Yes (Y/N)?");
    standard=answer();
    if (!standard) fprintf(fp,"#define MR_NO_STANDARD_IO\n");


    printf("Enter size of I/O buffer in bytes (1024 recommended) = ");
    scanf("%d",&iobsiz);
    getchar();
    if (iobsiz<128) iobsiz=128;
    fprintf(fp,"#define MR_IOBSIZ %d\n",iobsiz);

    if (!chosen)
    {
        printf("\n\nDo you for some reason NOT want to use a full-width number base?\n");
        printf("\nYou may not if your processor instruction set does not support\n");
        printf("%d-bit UNSIGNED multiply and divide instructions.\n",utlen);
        printf("If NOT then a full-width number base will be difficult and \n");
        printf("slow to implement, which is a pity, because its normally faster\n"); 
        printf("If for some other reason you don't want to use a full-width\n");
        printf("number base, (no muldvd()/muldvd2()/muldvm() available?), answer Yes\n");
        printf("\nAnswer (Y/N)?");
        nofull=answer();
        if (nofull) 
        {
            printf("\nRemember to use mirsys(...,MAXBASE), or somesuch, in your programs\n");
            printf("as mirsys(...,0); will generate an 'Illegal Number base' error\n");
            fprintf(fp,"#define MR_NOFULLWIDTH\n");
        }
    }
    if (port)
    {
          printf("\nDo you wish to use the Karatsuba/Comba/Montgomery method\n");
          printf("for modular arithmetic - as used by exponentiation\n");
          printf("cryptosystems like RSA. This method may be faster when\n");
          printf("using larger moduli, or if your processor has no \n");
          printf("unsigned integer multiply/divide instruction in its\n");
          printf("instruction set. This is true of some popular RISC\n");
          printf("computers.(Y/N)?");
          r=answer();
          if (r)
          {
              printf("\nThis method can only be used with moduli whose length in\n");
              printf("bits can be represented as %d*2^n, for any value\n",utlen*4);
              printf("of n. For example moduli of 256, 512, 1024 bits are OK\n");
              printf("\nSpecial fast routines for modular multiplication will now be\n");
              printf("automatically be invoked when, for example, powmod() is called\n");
              printf("\nRemember to use a full-width base in your programs\n");
              printf("by calling mirsys(..,0) or mirsys(..,256) at the start of the program\n");
              printf("Make sure to compile and link into your program the module MRKCM.C\n");
              fprintf(fp,"#define MR_KCM 4\n");
          }
    }

#ifdef __TURBOC__ 
    if (!port)
    {
      printf("\nDoes your computer have a Pentium processor\n");
      printf("and do you wish to exploit its built-in FP coprocessor\n");
      printf("for fastest possible modular exponentiation?\n");
      printf("NOTE: this is NOT optimal for Pentium Pro or Pentium II\n");
      printf("Supported only for 80x86 processors, and Borland C Compilers\n");
      printf("This is a little experimental - so use with care\n");
      printf("(Note however that although modular exponentiation will be fast,\n");
      printf("other operations will be relatively slow.) Answer (Y/N)?");
      r=answer();
      if (r) 
      {
          printf("Enter (maximum) modulus size in bits = ");
          scanf("%d",&nbits);
          getchar();
          b=31;
          do {
                  b--;
                  r=64-b-b;
                  s=1;
                  for (i=0;i<r;i++) s*=2;
                  s*=b;
          } while (s<=2*nbits);
          s=1; for (i=0;i<b;i++) s*=2;
          printf("\nDo you wish to generate variable length looping code, or\n");
          printf("fixed length unrolled code? The former can be used with any\n");
          printf("modulus less than the maximum size specified above. The latter will\n");
          printf("only work with a fixed modulus of that size, but is usually a bit\n");
          printf("faster, although it can generate a *lot* of code for larger moduli.\n"); 

          printf("\nAnswer Yes for looping code(Y/N)?");
          r=answer();
          if (r)
          {
           fprintf(fp,"#define MR_PENTIUM 0\n");
           printf("Make sure to compile and link into your program the module MR87V.C\n");
           }
          else
          {
           fprintf(fp,"#define MR_PENTIUM %d\n",nbits/b+1);
           printf("Make sure to compile and link into your program the module MR87F.C\n");
 
          }
           printf("\nSpecial fast routines for modular multiplication will now be\n");
           printf("automatically be invoked when, for example, powmod() is called\n");
           printf("\nIt is *vital* to use the appropriate number base, so\n");
           printf("you *must* now call mirsys(...,%ldL) at the start of your program\n",s);
      }
      else  if (!nofull)
      {
          printf("\nDo you wish to use the Karatsuba/Comba/Montgomery method\n");
          printf("for ultimate speed assembly language modular arithmetic\n");
          printf("specially with larger moduli\n");
          printf("Supported only for 80x86 processors, and Borland C\n");
          printf("compilers. Useful for exponentiation cryptosystems.(Y/N)?");
          r=answer();
          if (r)
          {
            printf("\nThis method can only be used with moduli whose length in\n");
            printf("bits can be represented as %d*(step size)*2^n, for any value\n",utlen);
            printf("of n. For example if you input a step size of 8, then \n");
            printf("moduli of 256, 512, 1024 bits etc will use this fast code\n");
            printf("The best step size can be determined by experiment, but\n");
            printf("larger step sizes generate more code. For the Pentium 8 is \n"); 
            printf("optimal. For the Pentium Pro/Pentium II 16 is optimal.\n"); 
            printf("If in doubt, set to 8\n"); 
            printf("Enter step size = ");
            scanf("%d",&words);
            getchar();
            fprintf(fp,"#define MR_KCM %d\n",words);
            printf("\nSpecial fast routines for modular multiplication will now be\n");
            printf("automatically be invoked when, for example, powmod() is called\n");
            printf("\nRemember to use a full-width base in your programs\n");
            printf("by calling mirsys(..,0) or mirsys(..,256) at the start of the program\n");
            printf("Make sure to compile and link into your program the module MRKCM.C\n");
          }
          else
          {
           printf("\nDo you want to create a 'comba' fixed size modular\n");
           printf("multiplier, for faster modular multiplication with\n");
           printf("smaller moduli\n");
           printf("Supported only for 80x86 processors, and Borland C\n");
           printf("compilers. Generates a *lot* of code \n");
           printf("Useful particularly for Elliptic Curve cryptosystems. (Y/N)?");
           r=answer();
           if (r)
           {
             printf("Enter modulus size in bits = ");
             scanf("%d",&nbits);
             getchar();
             words=nbits/utlen;
             if ((nbits%utlen)!=0) words++;
             fprintf(fp,"#define MR_COMBA %d\n",words);
             printf("Special fast routines for modular multiplication will now \n");
             printf("automatically be invoked when, for example, powmod() is called\n");
             printf("\nRemember to use a full-width base in your programs\n");
             printf("by calling mirsys(..,0) or mirsys(,..,256) at the start of the program\n");
             printf("Make sure to compile and link into your program the module MRCOMBA.C\n");
           }
          }
      }
    }
#endif

    fclose(fp);
    if (!port)
    {
        printf("\nYou must now provide an assembly language file mrmuldv.c,\n");
        if (!nofull)
        printf("containing implementations of muldiv(), muldvd(), muldvd2() and muldvm()\n");
        else printf("containing an implementation of muldiv()\n");
        printf("Check mrmuldv.any - an assembly language version may be\n");
        printf("there already, which can handle temporary %d bit values\n",dllen);
        printf("that arise from the multiplication of two %d bit quantities.\n",utlen); 
    }
    printf("\nA file mirdef.tst has been generated. If you are happy with it,\n");
    printf("rename it to mirdef.h and use for compiling the MIRACL library.\n");
}

