/*
 *   MIRACL memory allocation routines 
 *   mralloc.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 *
 *   MIRACL C Memory allocation/deallocation
 *   Can be replaced with special user-defined routines
 *   Default is to standard system routines
 *
 *   NOTE: uses calloc() which initialises memory to Zero, so make sure
 *   any substituted routine does the same!
 */

#include <stdio.h>
#include <miracl.h>
#include <stdlib.h>

extern miracl *mr_mip;

void mr_set_align(int n)
{ /* ad-hoc method to force alignment to n byte boundary. */
  /* Use with care - if at all */
    mr_mip->alignment=n;
}

void *mr_alloc(int num,int size)
{
    char *p; 
    int gap;

    if (mr_mip==NULL) 
    {
        p=calloc(num,size);
        if (p==NULL) mr_berror(MR_ERR_OUT_OF_MEMORY);
        return (void *)p;
    }
    if (!mr_mip->alignment) 
    {
        p=calloc(num,size);
        if (p==NULL) mr_berror(MR_ERR_OUT_OF_MEMORY);
        return (void *)p;
    }
    p=calloc(num*size+mr_mip->alignment,1);       /* Allocate a bit more */
    if (p==NULL) 
    {         
        mr_berror(MR_ERR_OUT_OF_MEMORY);
        return NULL;
    }
    gap=mr_mip->alignment-((unsigned long)p)%mr_mip->alignment;   
                                                  /* Move pointer forward */
    p+=gap;                                       /* to alignment boundary */
    *(p-1)=(unsigned char)gap;                    /* Remember increment */
    return (void *)p;
}

void mr_free(void *addr)
{
    char *p;
    int gap;
    if (addr==NULL) return;
    if (mr_mip==NULL)
    {
        free(addr);
        return;
    }
    if (!mr_mip->alignment) 
    {
        free(addr);
        return;
    }
    p=(char *)addr;
    gap=*(p-1);                    /* recall increment */
    p-=gap;                        /* step back to original address */
    free((void*)p);
}

