/*
 *   MIRACL arithmetic routines 3 - simple powers and roots
 *   mrarth3.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>
#define abs(x)  ((x)<0? (-(x)) : (x))

extern miracl *mr_mip;

int logb2(big x)
{ /* returns number of bits in x */
    int xl,lg2;
    if (mr_mip->ERNUM || size(x)==0) return 0;

    MR_IN(49)

    copy(x,mr_mip->w0);
    insign(PLUS,mr_mip->w0);
    lg2=0;
    xl=(int)mr_mip->w0[0];
    if (mr_mip->base==mr_mip->base2)
    {
        lg2=mr_mip->lg2b*(xl-1);
        mr_shift(mr_mip->w0,1-xl,mr_mip->w0);
    }
    else while (mr_mip->w0[0]>1)
    {
        mr_sdiv(mr_mip->w0,mr_mip->base2,mr_mip->w0);
        lg2+=mr_mip->lg2b;
    }
    while (mr_mip->w0[1]>=1)
    {
        lg2++;
        mr_mip->w0[1]/=2;
    }
    MR_OUT
    return lg2;
}

void expint(int b,int n,big x)
{ /* sets x=b^n */
    int i,r,p;
    unsigned int bit,un;
    if (mr_mip->ERNUM) return;
    convert(1,x);
    if (n==0) return;

    MR_IN(50)

    if (n<0)
    {
        mr_berror(MR_ERR_NEG_POWER);
        MR_OUT
        return;
    }
    if (b==2)
    {
        r=n/mr_mip->lg2b;
        p=n%mr_mip->lg2b;
        if (mr_mip->base==mr_mip->base2)
        {
            mr_shift(x,r,x);
            x[x[0]]<<=p;
        }
        else
        {
            for (i=1;i<=r;i++)
                mr_pmul(x,mr_mip->base2,x);
            mr_pmul(x,((mr_small)1<<p),x);
        }
    }
    else
    {
        bit=1;
        un=(unsigned int)n;
        while (un>=bit) bit<<=1;
        bit>>=1;
        while (bit>0)
        { /* ltr method */
            multiply(x,x,x);
            if ((bit&un)!=0) premult(x,b,x);
            bit>>=1;
        }
    }
    MR_OUT
}   

void sftbit(big x,int n,big z)
{ /* shift x by n bits */
    int m;
    if (mr_mip->ERNUM) return;
    copy(x,z);
    if (n==0) return;

    MR_IN(47)

    m=abs(n);
    if (n>0)
    { /* shift left */
        if (mr_mip->base==mr_mip->base2)
        {
            mr_shift(z,n/mr_mip->lg2b,z);
            mr_pmul(z,(mr_small)1<<(m%mr_mip->lg2b),z);
        }
        else
        {
            expint(2,m,mr_mip->w1);
            multiply(z,mr_mip->w1,z);
        }
    }
    else
    { /* shift right */
        if (mr_mip->base==mr_mip->base2)
        {
            mr_shift(z,n/mr_mip->lg2b,z);
            mr_sdiv(z,(mr_small)1<<(m%mr_mip->lg2b),z);
        }
        else
        {
            expint(2,m,mr_mip->w1);
            divide(z,mr_mip->w1,z);
        }
    }
    MR_OUT
}

void power(big x,long n,big z,big w)
{ /* raise big number to int power  w=x^n *
   * (mod z if z and w distinct)          */
    mr_small norm;
    copy(x,mr_mip->w5);
    zero(w);
    if(mr_mip->ERNUM || size(mr_mip->w5)==0) return;
    convert(1,w);
    if (n==0L) return;

    MR_IN(17)

    if (n<0L)
    {
        mr_berror(MR_ERR_NEG_POWER);
        MR_OUT
        return;
    }
    if (w==z) forever
    { /* "Russian peasant" exponentiation */
        if (n%2!=0L) multiply(w,mr_mip->w5,w);
        n/=2L;
        if (mr_mip->ERNUM || n==0L) break;
        multiply(mr_mip->w5,mr_mip->w5,mr_mip->w5);
    }
    else
    { 
        norm=normalise(z,z);
        divide(mr_mip->w5,z,z);
        forever
        {
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            if (n%2!=0L) mad(w,mr_mip->w5,mr_mip->w5,z,z,w);
            n/=2L;
            if (mr_mip->ERNUM || n==0L) break;
            mad(mr_mip->w5,mr_mip->w5,mr_mip->w5,z,z,mr_mip->w5);
        }
        if (norm!=1)
        {
            mr_sdiv(z,norm,z);
            divide(w,z,z);
        }
    }
    MR_OUT
}

BOOL nroot(big x,int n,big w)
{  /*  extract  lower approximation to nth root   *
    *  w=x^(1/n) returns TRUE for exact root      *
    *  uses Newtons method                        */
    int sx,dif,s,p,d,lg2,lgx,rem;
    BOOL full;
    if (mr_mip->ERNUM) return FALSE;
    if (size(x)==0 || n==1)
    {
        copy(x,w);
        return TRUE;
    }

    MR_IN(16)

    if (n<1) mr_berror(MR_ERR_BAD_ROOT);
    sx=exsign(x);
    if (n%2==0 && sx==MINUS) mr_berror(MR_ERR_NEG_ROOT);
    if (mr_mip->ERNUM) 
    {
        MR_OUT
        return FALSE;
    }
    insign(PLUS,x);
    lgx=logb2(x);
    if (n>=lgx)
    { /* root must be 1 */
        insign(sx,x);
        convert(sx,w);
        MR_OUT
        if (lgx==1) return TRUE;
        else        return FALSE;
    }
    expint(2,1+(lgx-1)/n,mr_mip->w2);           /* guess root as 2^(log2(x)/n) */
    s=(-(((int)x[0]-1)/n)*n);
    mr_shift(mr_mip->w2,s/n,mr_mip->w2);
    lg2=logb2(mr_mip->w2)-1;
    full=FALSE;
    if (s==0) full=TRUE;
    d=0;
    p=1;
    while (!mr_mip->ERNUM)
    { /* Newtons method */
        copy(mr_mip->w2,mr_mip->w3);
        mr_shift(x,s,mr_mip->w4);
        mr_mip->check=OFF;
        power(mr_mip->w2,n-1,mr_mip->w6,mr_mip->w6);
        mr_mip->check=ON;
        divide(mr_mip->w4,mr_mip->w6,mr_mip->w2);
        rem=size(mr_mip->w4);
        subtract(mr_mip->w2,mr_mip->w3,mr_mip->w2);
        dif=size(mr_mip->w2);
        subdiv(mr_mip->w2,n,mr_mip->w2);
        add(mr_mip->w2,mr_mip->w3,mr_mip->w2);
        p*=2;
        if(p<lg2+d*mr_mip->lg2b) continue;
        if (full && abs(dif)<n)
        { /* test for finished */
            while (dif<0)
            {
                rem=0;
                decr(mr_mip->w2,1,mr_mip->w2);
                mr_mip->check=OFF;
                power(mr_mip->w2,n,mr_mip->w6,mr_mip->w6);
                mr_mip->check=ON;
                dif=compare(x,mr_mip->w6);
            }
            copy(mr_mip->w2,w);
            insign(sx,w);
            insign(sx,x);
            MR_OUT
            if (rem==0 && dif==0) return TRUE;
            else                  return FALSE;
        }
        else
        { /* adjust precision */
            d*=2;
            if (d==0) d=1;
            s+=d*n;
            if (s>=0)
            {
                d-=s/n;
                s=0;
                full=TRUE;
            }
            mr_shift(mr_mip->w2,d,mr_mip->w2);
        }
        p/=2;
    }
    MR_OUT
    return FALSE;
}

