/*
 *   MIRACL flash trig.
 *   mrflsh3.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <math.h>
#include <miracl.h>

extern miracl *mr_mip;

#ifdef MR_FLASH

#define TAN 1
#define SIN 2
#define COS 3

static int norm(int type,flash y)
{ /* convert y to first quadrant angle *
   * and return sign of result         */
    int s;
    if (mr_mip->ERNUM) return 0;
    s=PLUS;
    if (size(y)<0)
    {
        negate(y,y);
        if (type!=COS) s=(-s);
    }
    fpi(mr_mip->pi);
    fpmul(mr_mip->pi,1,2,mr_mip->w8);
    if (fcomp(y,mr_mip->w8)<=0) return s;
    fpmul(mr_mip->pi,2,1,mr_mip->w8);
    if (fcomp(y,mr_mip->w8)>0)
    { /* reduce mod 2.pi */
        fdiv(y,mr_mip->w8,mr_mip->w9);
        ftrunc(mr_mip->w9,mr_mip->w9,mr_mip->w9);
        fmul(mr_mip->w9,mr_mip->w8,mr_mip->w9);
        fsub(y,mr_mip->w9,y);
    }
    if (fcomp(y,mr_mip->pi)>0)
    { /* if greater than pi */
        fsub(y,mr_mip->pi,y);
        if (type!=TAN) s=(-s);
    }
    fpmul(mr_mip->pi,1,2,mr_mip->w8);
    if (fcomp(y,mr_mip->w8)>0)
    { /* if greater than pi/2 */
        fsub(mr_mip->pi,y,y);
        if (type!=SIN) s=(-s);
    }
    return s;
}

static int tan1(big w,int n)
{  /* generator for C.F. of tan(1) */ 
    if (n==0) return 1;
    if (n%2==1) return 2*(n/2)+1;
    else return 1;
}

void ftan(flash x,flash y)
{ /* calculates y=tan(x) */
    int i,n,nsq,m,sqrn,sgn,op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(57)

    sgn=norm(TAN,y);
    ftrunc(y,y,mr_mip->w10);
    n=size(y);
    if (n!=0) build(y,tan1);
    if (size(mr_mip->w10)==0)
    {
        insign(sgn,y);
        MR_OUT
        return;
    }
    sqrn=isqrt(mr_mip->lg2b*mr_mip->workprec,mr_mip->lg2b);
    nsq=0;
    copy(mr_mip->w10,mr_mip->w8);
    frecip(mr_mip->w10,mr_mip->w10);
    ftrunc(mr_mip->w10,mr_mip->w10,mr_mip->w10);
    m=logb2(mr_mip->w10);
    if (m<sqrn)
    { /* scale fraction down */
        nsq=sqrn-m;
        expint(2,nsq,mr_mip->w10);
        fdiv(mr_mip->w8,mr_mip->w10,mr_mip->w8);
    }
    zero(mr_mip->w10);
    fmul(mr_mip->w8,mr_mip->w8,mr_mip->w9);
    negate(mr_mip->w9,mr_mip->w9);
    op[0]=0x4B;    /* set up for x/(y+C) */
    op[1]=op[3]=1;
    op[2]=0;
    for (m=sqrn;m>1;m--)
    { /* Unwind C.F for tan(x)=z/(1-z^2/(3-z^2/(5-...)))))) */
        op[4]=2*m-1;
        flop(mr_mip->w9,mr_mip->w10,op,mr_mip->w10);
    }
    op[4]=1;
    flop(mr_mip->w8,mr_mip->w10,op,mr_mip->w10);
    op[0]=0x6C;     /* set up for tan(x+y)=tan(x)+tan(y)/(1-tan(x).tan(y)) */
    op[1]=op[2]=op[3]=1;
    op[4]=(-1);
    for (i=0;i<nsq;i++)
    { /* now square it back up again */
        flop(mr_mip->w10,mr_mip->w10,op,mr_mip->w10);
    }
    flop(y,mr_mip->w10,op,y);
    insign(sgn,y);
    MR_OUT
}

void fatan(flash x,flash y)
{ /* calculate y=atan(x) */
    int s,c,op[5];
    BOOL inv,hack;
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(58)

    s=exsign(y);
    insign(PLUS,y);
    inv=FALSE;
    fpi(mr_mip->pi);
    fconv(1,1,mr_mip->w11);
    c=fcomp(y,mr_mip->w11);
    if (c==0)
    { /* atan(1)=pi/4 */
        fpmul(mr_mip->pi,1,4,y);
        insign(s,y);
        MR_OUT
        return;
    }
    if (c>0)
    { /* atan(x)=pi/2 - atan(1/x) for x>1 */
        inv=TRUE;
        frecip(y,y);
    }
    hack=FALSE;
    if (mr_lent(y)<=2)
    { /* for 'simple' values of y */
        hack=TRUE;
        fconv(3,1,mr_mip->w11);
        froot(mr_mip->w11,2,mr_mip->w11);
        op[0]=0xC6;
        op[2]=op[3]=op[4]=1;
        op[1]=(-1);
        flop(y,mr_mip->w11,op,y);
    }
    op[0]=0x4B;
    op[1]=op[3]=op[4]=1;
    op[2]=0;
    mr_mip->workprec=mr_mip->stprec;
    dconv(atan(fdsize(y)),mr_mip->w11);
    while (mr_mip->workprec!=mr_mip->nib)
    { /* Newtons iteration w11=w11+(y-tan(w11))/(tan(w11)^2+1) */
        if (mr_mip->workprec<mr_mip->nib) mr_mip->workprec*=2;
        if (mr_mip->workprec>=mr_mip->nib) mr_mip->workprec=mr_mip->nib;
        else if (mr_mip->workprec*2>mr_mip->nib) mr_mip->workprec=(mr_mip->nib+1)/2;
        ftan(mr_mip->w11,mr_mip->w12);
        fsub(y,mr_mip->w12,mr_mip->w8);
        fmul(mr_mip->w12,mr_mip->w12,mr_mip->w12);
        flop(mr_mip->w8,mr_mip->w12,op,mr_mip->w12);  /* w12=w8/(w12+1) */
        fadd(mr_mip->w12,mr_mip->w11,mr_mip->w11);
    }
    copy(mr_mip->w11,y);
    op[0]=0x6C;
    op[1]=op[3]=6;
    op[2]=1;
    op[4]=0;
    if (hack) flop(y,mr_mip->pi,op,y);
    op[1]=(-2);
    op[3]=2;
    if (inv) flop(y,mr_mip->pi,op,y);
    insign(s,y);
    MR_OUT
}

void fsin(flash x,flash y)
{ /*  calculate y=sin(x) */
    int sgn,op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(59)
    sgn=norm(SIN,y);
    fpmul(y,1,2,y);
    ftan(y,y);
    op[0]=0x6C;
    op[1]=op[2]=op[3]=op[4]=1;
    flop(y,y,op,y);
    insign(sgn,y);
    MR_OUT
}

void fasin(flash x,flash y)
{ /* calculate y=asin(x) */
    int s,op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;
  
    MR_IN(60)
    s=exsign(y);
    insign(PLUS,y);
    op[0]=0x3C;
    op[1]=(-1);
    op[2]=op[3]=1;
    op[4]=0;
    flop(y,y,op,mr_mip->w11);  /* mr_w11 = -(y.y-1) */
    froot(mr_mip->w11,2,mr_mip->w11);
    if (size(mr_mip->w11)==0)
    {
        fpi(mr_mip->pi);
        fpmul(mr_mip->pi,1,2,y);
    }
    else
    {
        fdiv(y,mr_mip->w11,y);
        fatan(y,y);
    }
    insign(s,y);    
    MR_OUT
}

void fcos(flash x,flash y)
{ /* calculate y=cos(x) */
    int sgn,op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0)
    {
        convert(1,y);
        return;
    }

    MR_IN(61)
    sgn=norm(COS,y);
    fpmul(y,1,2,y);
    ftan(y,y);
    op[0]=0x33;
    op[1]=op[3]=op[4]=1;
    op[2]=(-1);
    flop(y,y,op,y);
    insign(sgn,y);
    MR_OUT
}

void facos(flash x,flash y)
{ /* calculate y=acos(x) */
    int op[5];
    if (mr_mip->ERNUM) return;

    MR_IN(62)

    fasin(x,y);
    fpi(mr_mip->pi);
    op[0]=0x6C;
    op[1]=(-2);
    op[2]=1;
    op[3]=2;
    op[4]=0;
    flop(y,mr_mip->pi,op,y);    /* y = pi/2 - y */
    MR_OUT
}

#endif

