/*
 *   MIRACL flash hyperbolic trig.
 *   mrflsh4.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

extern miracl *mr_mip;

#ifdef MR_FLASH

void ftanh(flash x,flash y)
{ /* calculates y=tanh(x) */
    int op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(63)
    fexp(y,y);
    op[0]=0x33;
    op[1]=op[3]=op[4]=1;
    op[2]=(-1);
    flop(y,y,op,y);
    MR_OUT
}

void fatanh(flash x,flash y)
{ /* calculate y=atanh(x) */
    int op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(64)
    fconv(1,1,mr_mip->w11);
    op[0]=0x66;
    op[1]=op[2]=op[3]=1;
    op[4]=(-1);
    flop(mr_mip->w11,y,op,y);
    flog(y,y);
    fpmul(y,1,2,y);
    MR_OUT
}

void fsinh(flash x,flash y)
{ /*  calculate y=sinh(x) */
    int op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(65)
    fexp(y,y);
    op[0]=0xC6;
    op[2]=op[3]=op[4]=1;
    op[1]=(-1);
    flop(y,y,op,y);
    MR_OUT
}

void fasinh(flash x,flash y)
{ /* calculate y=asinh(x) */
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0) return;

    MR_IN(66)
    fmul(y,y,mr_mip->w11);
    fincr(mr_mip->w11,1,1,mr_mip->w11);
    froot(mr_mip->w11,2,mr_mip->w11);
    fadd(y,mr_mip->w11,y);
    flog(y,y);
    MR_OUT
}

void fcosh(flash x,flash y)
{ /* calculate y=cosh(x) */
    int op[5];
    copy(x,y);
    if (mr_mip->ERNUM || size(y)==0)
    {
        convert(1,y);
        return;
    }

    MR_IN(67)
    fexp(y,y);
    op[0]=0xC6;
    op[1]=op[2]=op[3]=op[4]=1;
    flop(y,y,op,y);
    MR_OUT
}

void facosh(flash x,flash y)
{ /* calculate y=acosh(x) */
    copy(x,y);
    if (mr_mip->ERNUM) return;

    MR_IN(62)
    fmul(y,y,mr_mip->w11);
    fincr(mr_mip->w11,(-1),1,mr_mip->w11);
    froot(mr_mip->w11,2,mr_mip->w11);
    fadd(y,mr_mip->w11,y);
    flog(y,y);
    MR_OUT
}

#endif

