/*
 *   MIRACL I/O routines 1. 
 *   mrio1.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

extern miracl *mr_mip;

int instr(flash x,char *string)
{  /*  input a big number       *
    *  returns length in digits */
    int i,ipt,n,s,e;
    unsigned int ch;
#ifdef MR_FLASH
    BOOL frac;
#endif
    if (mr_mip->ERNUM) return 0;

    MR_IN(76)

    if (mr_mip->apbase==0 || mr_mip->apbase>256)
    {
        mr_berror(MR_ERR_BASE_TOO_BIG);
        MR_OUT
        return 0;
    }
    if (!mr_mip->active)
    {
        mr_berror(MR_ERR_NO_MIRSYS);
        MR_OUT
        return 0;
    }
    zero(x);
    if (mr_mip->fin) string=mr_mip->IOBUFF;
    if (mr_mip->INPLEN==0)
    { /* inputting ASCII bytes */
        if (mr_mip->fin)
        { /* read in characters */
            i=0;
            do
            {
                ch=(unsigned char)fgetc(mr_mip->infile);
                if (ch==EOF) break;
                string[i++]=ch;
                if (i>=MR_IOBSIZ)
                {
                    mr_berror(MR_ERR_IO_OVERFLOW);
                    MR_OUT
                    return 0;
                }
            } while (ch!='\n' && ch!='\0'); 
            string[i]='\0';
        }
        forever
        { /*  get input length  */
            ch=(unsigned char)string[mr_mip->INPLEN];
            if (ch=='\0') break;
            if (mr_mip->apbase<=60 && ch=='\n') break;
            mr_mip->INPLEN++;
            if (string==mr_mip->IOBUFF && mr_mip->INPLEN>=MR_IOBSIZ)
            {
                mr_berror(MR_ERR_IO_OVERFLOW);
                MR_OUT
                return 0;
            }
        }
    }
    else
    { /* inputting BINARY bytes */
        if (string==mr_mip->IOBUFF && mr_mip->INPLEN>=MR_IOBSIZ)
        {
            mr_berror(MR_ERR_IO_OVERFLOW);
            MR_OUT
            return 0;
        }
        if (mr_mip->fin) for(i=0;i<mr_mip->INPLEN;i++) 
        {
            if ((ch=(unsigned char)fgetc(mr_mip->infile))==EOF)
            {
                mr_mip->INPLEN=i;
                break;
            }
            string[i]=(char)ch;
        }
    }
    n=0;
    s=PLUS;
    e=0;
#ifdef MR_FLASH
    frac=FALSE;
#endif
    if (mr_mip->INPLEN>0 && mr_mip->apbase<=60)
    { /* skip leading blanks and check sign */
#ifdef MR_FLASH
        if (string[mr_mip->INPLEN-1]=='/') mr_mip->INPLEN--;
#endif
        while (string[e]==' ') e++;
        if (string[e]=='-')
        { /* check sign */
             s=MINUS;
             e++;
        }
        if (string[e]=='+') e++;
    }
    for (i=mr_mip->INPLEN-1;i>=e;i--)
    {
        ch=(unsigned char)string[i];
        if (mr_mip->apbase<=60)
        { /* check for slash or dot and convert character to number */
#ifdef MR_FLASH
            if (!frac)
            {
                if (ch=='/')
                {
                    frac=TRUE;
                    copy(x,mr_mip->w0);
                    zero(x);    
                    n=0;
                    continue;
                }
                if (ch=='.')
                {
                    frac=TRUE;
                    zero(mr_mip->w0);
                    putdig(1,mr_mip->w0,n+1);
                    continue;
                }
            }
#endif
            ch+=80;
            if (ch>127 && ch<138) ch-=128;
            if (ch>144 && ch<171) ch-=135;
            if (ch>176 && ch<203) ch-=141;
        }
        if ((mr_small)ch>=mr_mip->apbase)
        {
            mr_berror(MR_ERR_BAD_FORMAT);
            MR_OUT
            return 0;
        }
        n++;
        putdig(ch,x,n);
    }
    ipt=mr_mip->INPLEN;
    mr_mip->INPLEN=0;
    insign(s,x);
    mr_lzero(x);
#ifdef MR_FLASH
    mr_lzero(mr_mip->w0);
    if (frac) fpack(x,mr_mip->w0,x);
#endif
    MR_OUT
    return ipt;
}

int otstr(flash x,char *string)
{  /*  output a big number  */
    int s,i,n,ch,rp,nd,m;
#ifdef MR_FLASH
    int nw,dw;
#endif
    mr_small lx;
    BOOL done;
    if (mr_mip->ERNUM) return 0;

    MR_IN(75)

    if (mr_mip->apbase==0 || mr_mip->apbase>256)
    {
        mr_berror(MR_ERR_BASE_TOO_BIG);
        MR_OUT
        return 0;
    }
    if (!mr_mip->active)
    {
        mr_berror(MR_ERR_NO_MIRSYS);
        MR_OUT
        return 0;
    }
    n=0;
    s=exsign(x);
    insign(PLUS,x);
    lx = x[0];
    if (lx==0 && mr_mip->apbase<=60)
    {
        if (!mr_mip->fout)
        {
            string[0]='0';
            string[1]='\0';
        }
        else
        {
            fputc('0',mr_mip->otfile);
            fputc('\n',mr_mip->otfile);
        }
        MR_OUT
        return 1;
    }
    rp=0;
    if (s==MINUS && mr_mip->apbase<=60)
    {
        if (!mr_mip->fout) string[n]='-';
        else fputc('-',mr_mip->otfile);
        n++;
    }
#ifdef MR_FLASH
    done=FALSE;
    numer(x,mr_mip->w6);
    if (mr_mip->RPOINT)
    { /* output with radix point */
        denom(x,mr_mip->w5);
        if (size(mr_mip->w5)>1)
        { /* multiply up numerator to get full precision in *
           * the output. Remember position of radix point.  */
            nw=(int)(lx&mr_mip->MSK);
            dw=(int)((lx>>mr_mip->BTS)&mr_mip->MSK);
            if (nw==0) nw++;
            mr_mip->check=OFF;
            if (nw>dw) mr_shift(mr_mip->w5,nw-dw,mr_mip->w5);
            if (dw>nw) mr_shift(mr_mip->w6,dw-nw,mr_mip->w6);
            nd=mr_mip->nib;
            if (compare(mr_mip->w6,mr_mip->w5)>=0) nd--;
            copy(mr_mip->w6,mr_mip->w0);
            if (((int)mr_mip->w0[0]+nd)>2*mr_mip->nib) nd=2*mr_mip->nib-(int)mr_mip->w0[0];
            mr_shift(mr_mip->w0,nd,mr_mip->w0);
            divide(mr_mip->w0,mr_mip->w5,mr_mip->w6);
            mr_mip->check=ON;
            rp=mr_mip->pack*(nd+dw-nw);
        }
    }
#else
    copy(x,mr_mip->w6);
    done=TRUE;
#endif

    forever
    {
        nd=numdig(mr_mip->w6);
        m=nd;
        if (rp>m) m=rp;
        for (i=m;i>0;i--)
        { 
            if (!mr_mip->fout && string==mr_mip->IOBUFF && n>=MR_IOBSIZ-3)
            {
                mr_berror(MR_ERR_IO_OVERFLOW);
                MR_OUT
                return n;
            }
#ifdef MR_FLASH
            if (i==rp && mr_mip->apbase<=60)
            {
                if (!mr_mip->fout) string[n]='.';
                else fputc('.',mr_mip->otfile);
                n++;
            }
#endif
            if (i>nd) ch='0';
            else
            {
                ch=getdig(mr_mip->w6,i);
                mr_mip->check=OFF;
                putdig(0,mr_mip->w6,i);
                mr_mip->check=ON;
                if (mr_mip->apbase<=60)
                { /* convert number to character */
                    ch+=48;
                    if (ch>=58) ch+=7;
                    if (ch>=91) ch+=6;
                }
            }
            if (i<rp && mr_mip->apbase<=60 && ch=='0' && size(mr_mip->w6)==0) break;
            if (!mr_mip->fout) string[n]=(char)ch;
            else fputc((char)ch,mr_mip->otfile);
            n++;
        }
        if (done || mr_mip->RPOINT) break;
#ifdef MR_FLASH
        denom(x,mr_mip->w6);
        if (size(mr_mip->w6)==1) break;
        if (mr_mip->apbase<=60)
        {
            if (!mr_mip->fout) string[n]='/';
            else fputc('/',mr_mip->otfile);
            n++;
        }
        done=TRUE;
#endif
    }
/* Append a trailing 0 - it may be printable ascii text */    
    if (!mr_mip->fout) string[n]='\0';
    else if (mr_mip->apbase<=60) fputc('\n',mr_mip->otfile);
    insign(s,x);
    MR_OUT
    return n;
}

int innum(flash x,FILE *filep)
{ /* convert from string to flash x */
    int n;
    if (mr_mip->ERNUM) return 0;

    MR_IN(1)

    mr_mip->infile=filep;
    mr_mip->fin=TRUE;
    n=instr(x,NULL);
    mr_mip->fin=FALSE;

    MR_OUT
    return n;
}

int otnum(flash x,FILE *filep)
{ /* convert flash to string */
    int n;
    if (mr_mip->ERNUM) return 0;

    MR_IN(2)

    mr_mip->otfile=filep;
    mr_mip->fout=TRUE;
    n=otstr(x,NULL);
    mr_mip->fout=FALSE;

    MR_OUT
    return n;
}

