/*
 *   MIRACL I/O routines 2. 
 *   mrio2.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

extern miracl *mr_mip;

static void cbase(big x,mr_small oldbase,big y)
{  /*  change radix of x from oldbase to base  */
    int i,s;
    mr_small n;
    BOOL done;
    if (mr_mip->ERNUM) return;
    if (mr_mip->base==oldbase)
    {
        copy(x,y);
        return;
    }

    MR_IN(13)

    s=exsign(x);
#ifdef MR_FLASH
    numer(x,mr_mip->w1);
    denom(x,mr_mip->w2);
    done=FALSE;
#else
    copy(x,mr_mip->w1);
    done=TRUE;
#endif
    insign(PLUS,mr_mip->w1);
    forever
    {
        zero(mr_mip->w6);
        convert(1,mr_mip->w0);
        for (i=1;i<=(int)mr_mip->w1[0];i++)  
        { /* this is a bit slow - but not time critical */
            mr_pmul(mr_mip->w0,mr_mip->w1[i],mr_mip->w5);
            add(mr_mip->w6,mr_mip->w5,mr_mip->w6);
            if (oldbase==0)
            { /* bit of a frig! */
                n=((mr_small)1<<(MIRACL/2));
                mr_pmul(mr_mip->w0,n,mr_mip->w0);
                mr_pmul(mr_mip->w0,n,mr_mip->w0);
            }
            else mr_pmul(mr_mip->w0,oldbase,mr_mip->w0);
        }
        if (mr_mip->ERNUM || done) break;
#ifdef MR_FLASH
        copy(mr_mip->w2,mr_mip->w1);
        copy(mr_mip->w6,mr_mip->w7);
        done=TRUE;
#endif
    }
#ifdef MR_FLASH
    fpack(mr_mip->w7,mr_mip->w6,y);
#else
    copy(mr_mip->w6,y);
#endif
    insign(s,y);
    MR_OUT
}

int cinstr(flash x,char *string)
{  /*  input big number in base IOBASE  */
    mr_small newb,oldb,b,lx;
    int ipt;
    if (mr_mip->ERNUM) return 0;

    MR_IN(78)

    newb=mr_mip->IOBASE;
    oldb=mr_mip->apbase;
    mr_setbase(newb); /* temporarily change base ... */
    b=mr_mip->base;
    mr_mip->check=OFF;
    ipt=instr(mr_mip->w5,string); /* ... and get number  */
    mr_mip->check=ON;
    lx=(mr_mip->w5[0]&mr_mip->OBITS);
#ifdef MR_FLASH
    if ((int)(lx&mr_mip->MSK)>mr_mip->nib || (int)((lx>>mr_mip->BTS)&mr_mip->MSK)>mr_mip->nib)
#else
    if ((int)lx>mr_mip->nib)
#endif
    { /* numerator or denominator too big */
        mr_berror(MR_ERR_OVERFLOW);
        MR_OUT
        return 0;
    }
    mr_setbase(oldb);      /* restore original base */
    cbase(mr_mip->w5,b,x);
    MR_OUT
    return ipt;
}   

int cotstr(flash x,char *string)
{  /*  output a big number in base IOBASE  */
    mr_small newb,oldb,b;
    int ipt;
    if (mr_mip->ERNUM) return 0;

    MR_IN(77)
    newb=mr_mip->IOBASE;
    oldb=mr_mip->apbase;
    b=mr_mip->base;
    mr_setbase(newb);   /* temporarily change base ... */
    mr_mip->check=OFF;
    cbase(x,b,mr_mip->w5);   /* number may get bigger !     */   
    mr_mip->check=ON;
    ipt=otstr(mr_mip->w5,string); /*..... and output number  */ 
    zero(mr_mip->w5);
    mr_setbase(oldb);
    MR_OUT
    return ipt;
}

int cinnum(flash x,FILE *filep)
{ /* convert from string to flash x */
    int n;
    if (mr_mip->ERNUM) return 0;

    MR_IN(14)
    mr_mip->infile=filep;
    mr_mip->fin=TRUE;
    n=cinstr(x,NULL);
    mr_mip->fin=FALSE;
    MR_OUT
    return n;
}

int cotnum(flash x,FILE *filep)
{ /* convert flash to string */
    int n;
    if (mr_mip->ERNUM) return 0;

    MR_IN(15)
    mr_mip->otfile=filep;
    mr_mip->fout=TRUE;
    n=cotstr(x,NULL);
    mr_mip->fout=FALSE;
    MR_OUT
    return n;
}

