/*
 *   MIRACL Chinese Remainder Thereom routines (for use with small moduli) 
 *   mrscrt.c
 *
 *   Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <stdlib.h>
#include <miracl.h>

extern miracl *mr_mip;

BOOL scrt_init(small_chinese *c,int r,mr_utype *moduli)
{ /* calculate CRT constants - returns FALSE if there is a problem */
    int i,j,k;
    if (r<1) return FALSE;
    if (r==1)
    {
        c->NP=1;
        c->M=(mr_utype *)mr_alloc(r,sizeof(mr_utype));
        if (c->M==NULL) return FALSE;
        c->M[0]=moduli[0];
        return TRUE;
    }
    for (i=0;i<r;i++) if (moduli[i]<2) return FALSE;
    c->M=(mr_utype *)mr_alloc(r,sizeof(mr_utype));
    if (c->M==NULL) return FALSE;
    c->C=(mr_utype *)mr_alloc(r*(r-1)/2,sizeof(mr_utype));
    if (c->C==NULL)
    { /* no room */
        mr_free(c->M);
        return FALSE;
    }
    c->V=(mr_utype *)mr_alloc(r,sizeof(mr_utype));
    if (c->V==NULL)
    { /* no room */
        mr_free(c->M);
        mr_free(c->C);
        return FALSE;
    }
    for (k=0,i=0;i<r;i++)
    {
        c->M[i]=moduli[i];
        for (j=0;j<i;j++,k++)
            c->C[k]=invers(c->M[j],c->M[i]);
    }
    c->NP=r;
    return TRUE;
}

void scrt(small_chinese *c,mr_utype *u,big x)
{ /* Chinese Remainder Thereom                  *
   * Calculate x given remainders u[i] mod M[i] */
    int i,j,k;
    mr_utype *V,*C,*M;
    mr_small t;
    V=c->V;
    C=c->C;
    M=c->M;
    if (c->NP<1) return;
    if (c->NP==1)
    {
        t=smul(1,u[0],M[0]);
        convert(1,mr_mip->w1);
        mr_pmul(mr_mip->w1,t,x);
        return;
    }
    V[0]=u[0];
    for (k=0,i=1;i<c->NP;i++)
    { /* Knuth P. 274 */
        V[i]=u[i] - V[0];
        V[i]=smul(V[i],C[k],M[i]);
        k++;

#if INLINE_ASM == 3

        ASM mov ebx,DWORD PTR V
        ASM mov esi,DWORD PTR M
        ASM mov edi,DWORD PTR C
        ASM mov ecx,1
        ASM mov edx,DWORD PTR i
        ASM mov esi,[esi+4*edx]

/*
   ebx = V
   esi = M[i]
   edi = C
   ecx = j
   edx = i,k
*/
     s1:
        ASM cmp ecx,edx
        ASM jge s2

        ASM mov eax,[ebx+4*edx]
        ASM push edx

        ASM sub eax,[ebx+4*ecx]
        ASM cdq
        ASM idiv esi
        ASM mov eax,edx
        ASM add eax,esi

        ASM mov edx,DWORD PTR k
        ASM mul DWORD PTR [edi+4*edx]
        ASM div esi

        ASM mov eax,edx

        ASM pop edx
        ASM mov [ebx+4*edx],eax   
        ASM inc DWORD PTR k
        ASM inc ecx
        ASM jmp s1
     s2:
        ASM nop

#else
        for (j=1;j<i;j++,k++)
        {
            V[i]-=V[j];
            V[i]=smul(V[i],C[k],M[i]);
        }
#endif
    }

    convert(1,x);
    mr_pmul(x,(mr_small)V[0],x);
    convert(1,mr_mip->w1);
    for (i=1;i<c->NP;i++)
    {
        mr_pmul(mr_mip->w1,(mr_small)(M[i-1]),mr_mip->w1);
        mr_pmul(mr_mip->w1,(mr_small)(V[i]),mr_mip->w2);
        add(x,mr_mip->w2,x);         
    } 
}

void scrt_end(small_chinese *c)
{ /* clean up after CRT */
    if (c->NP<1) 
    {
        c->NP=0;
        return;
    }
    if (c->NP==1)
    {
        mr_free(c->M);
        c->NP=0;
        return;
    }
    mr_free(c->M);
    mr_free(c->V);
    mr_free(c->C);
    c->NP=0;
}

