/*
 *   MIRACL cryptographic strong random number generator 
 *   mrstrong.c
 *
 *   Copyright (c) 1988-1998 Shamus Software Ltd.
 *
 *   Unguessable seed -> SHA -> PRNG internal state -> SHA -> random numbers
 *   Slow - but secure
 *
 *   See http://www.rsa.com/rsalabs/pubs/updates/bull-1.pdf for a justification
 */

#include <stdio.h>
#include <miracl.h>

extern miracl *mr_mip;

static void fill_pool(void)
{ /* hash down output of RNG to re-fill the pool */
    int i;
    sha sh;
    shs_init(&sh);
    for (i=0;i<128;i++) shs_process(&sh,(int)brand());
    shs_hash(&sh,mr_mip->pool);
    mr_mip->pool_ptr=0;
}

void strong_init(int rawlen,char *raw,mr_unsign32 tod)
{ /* initialise from at least 128 byte string of raw  *
   * random (keyboard?) input, and 32-bit time-of-day */
    int i;
    mr_unsign32 hash[MR_HASH_BYTES/4];
    sha sh;
    mr_mip->pool_ptr=0;
    shs_init(&sh);
    for (i=0;i<rawlen;i++)
        shs_process(&sh,raw[i]);
    shs_hash(&sh,(char *)hash);

/* initialise PRNG from distilled randomness */
    irand(tod);
    for (i=0;i<MR_HASH_BYTES/4;i++) irand(hash[i]);

    fill_pool();

}

void strong_rng(int bytes,char *ran)
{ 
    int i;
    for (i=0;i<bytes;i++)
    {
        ran[i]=mr_mip->pool[mr_mip->pool_ptr++];
        if (mr_mip->pool_ptr>=MR_HASH_BYTES) fill_pool();
    }    
}

/* test main program 

#include <stdio.h>
#include <miracl.h>
#include <time.h>

void main()
{
    char raw[256],bytes[25];
    big x;
    long seed;
    miracl *mip=mirsys(50,256);
    x=mirvar(0);
    printf("Enter Raw random string= ");
    scanf("%s",raw);
    getchar();
    time(&seed);
    strong_init(strlen(raw),raw,seed);
    mip->IOBASE=16;
    forever
    {
        strong_rng(25,bytes);
        mip->INPLEN=25;
        instr(x,bytes);
        cotnum(x,stdout);
    }
}

*/

