/*
 *   Cramer-Shoup Public Key Encryption - Hybrid Implementation 
 *
 *   This program encrypts a file fred.xxx to fred.crs. The AES encryption
 *   key is encrypted under the public key in the file fred.key
 *
 *   The public key comes from from public.crs
 *
 *   Copyright (c) 1999 Shamus Software Ltd.
 */

#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <big.h>

Miracl precision(300,256);

void strip(char *name)
{ /* strip off filename extension */
    int i;
    for (i=0;name[i]!='\0';i++)
    {
        if (name[i]!='.') continue;
        name[i]='\0';
        break;
    }
}

void hash(sha *s,Big x)
{ /* hash in a big number one byte at a time */
    sha sh;
    int ch;
    while (x!=0)
    {
        ch=x%256; x/=256;
        shs_process(s,ch);
    }
}

int main()
{
    int i;
    ifstream common("common.crs");    // construct file I/O streams 
    ifstream plaintext,public_key("public.crs");
    ofstream key_file,ciphertext;
    Big p,q,g1,g2,h,m,r,u1,u2,e,v,ex,c,d,alpha;
    static char ifname[100],ofname[100];
    char ch,H[20],iv[16],block[16]; 
    sha sh;
    aes a;
    long seed;
    miracl *mip=&precision;

// randomise 
    cout << "Enter 9 digit random number seed  = ";
    cin >> seed;
    irand(seed);
    mip->IOBASE=16;

// get common data 
    common >> q >> p >> g1 >> g2;

// get public key of signer 
    public_key >> c >> d >> h;

    r=rand(q);

    shs_init(&sh);
    hash(&sh,pow(h,r,p));                          
    shs_hash(&sh,H);      // AES Key = Hash(h^r mod p)
    for (i=0;i<16;i++) iv[i]=i;  // Set CBC IV
    aes_init(&a,16,H,iv);        // Set 128 bit AES Key

// figure out where input is coming from

    cout << "file to be encoded = " ;
    cin >> ifname;

   /* set up input file */
    strcpy(ofname,ifname);
    strip(ofname);
    strcat(ofname,".crs");
    plaintext.open(ifname); 
    if (!plaintext)
    {
        cout << "Unable to open file " << ifname << "\n";
        return 0;
    }
    cout << "encoding message\n";

    ciphertext.open(ofname,ios::binary);

    u1=pow(g1,r,p);
    u2=pow(g2,r,p);

    shs_init(&sh);
    hash(&sh,u1);
    hash(&sh,u2);

// now encrypt the plaintext 
// and hash in the ciphertext.

    int bptr=0;
    forever
    { // encrypt input ..
        plaintext.get(ch);
        if (plaintext.eof()) break;
        block[bptr++]=ch;          // collect one character
        if (bptr==16)
        { // full block ready - encrypt and output
            bptr=0;
            aes_encrypt(&a,block);
            for (i=0;i<16;i++) 
            { // include ciphertext in hash
                shs_process(&sh,block[i]);
                ciphertext << block[i];
            }
        } 
    }

// pad last block with 0's
    if (bptr>0)
    { // something left over...
        for (i=bptr;i<16;i++) block[i]=0;
        aes_encrypt(&a,block);
        for (i=0;i<16;i++) 
        {
            shs_process(&sh,block[i]);
            ciphertext << block[i];
        }
    }

    aes_end(&a);

    shs_hash(&sh,H);
    mip->INPLEN=20; 
    mip->IOBASE=256;
    alpha=(char *)H;   // alpha=Hash(u1,u2,ciphertext)

    ex=(r*alpha)%q;
    v=pow(c,r,d,ex,p);      // multi-exponentiation

    strip(ofname);
    strcat(ofname,".key");
    mip->IOBASE=16;
    key_file.open(ofname);
    key_file << u1 << endl;
    key_file << u2 << endl;
    key_file << v << endl;
    return 0;
}

