/*
 * C++ class to implement a polynomial type and to allow 
 * arithmetic on polynomials whose elements are from
 * the finite field mod p
 *
 * WARNING: This class has been cobbled together for a specific use with
 * the MIRACL library. It is not complete, and may not work in other 
 * applications
 *
 * See Knuth The Art of Computer Programming Vol.2, Chapter 4.6 
 */

#include "poly.h"

Poly::Poly(const Poly& p)
{
    term *ptr=p.start;
    start=NULL;
    while (ptr!=NULL)
    {  
        addterm(ptr->an,ptr->n);
        ptr=ptr->next;
    }    
}

Poly::~Poly()
{
   term *nx;
   while (start!=NULL)
   {   
       nx=start->next;
       delete start;
       start=nx;
   }
}


ZZn Poly::coeff(int power)
{
    ZZn c=0;
    term *ptr=start;
    while (ptr!=NULL)
    {
        if (ptr->n==power)
        {
            c=ptr->an;
            return c;
        }
        ptr=ptr->next;
    }
    return c;
}

ZZn Poly::F(const ZZn& x)
{
    ZZn f=0;
    term *ptr=start;
    while (ptr!=NULL)
    {
        f+=(ptr->an*pow(x,ptr->n));  // add in an*x^n
        ptr=ptr->next;
    }
    return f;
}

ZZn Poly:: min()
{
    term *ptr=start;
    if (start==NULL) return (ZZn)0;
    
    while (ptr->next!=NULL) ptr=ptr->next;
    return (ptr->an);
}

Poly compose(const Poly& g,const Poly& b,const Poly& m)
{ // compose polynomials
  // assume G(x) = G3x^3 + G2x^2 + G1x^1 +G0
  // Calculate G(B(x) = G3.(B(x))^3 + G2.(B(x))^2 ....   
    Poly c,t;  
    term *ptr;
    int i,d=degree(g);
    Poly *table=new Poly[d+1];
    table[0].addterm((ZZn)1,0);
    for (i=1;i<=d;i++) table[i]=(table[i-1]*b)%m;
    ptr=g.start;
    while (ptr!=NULL)
    {
        c+=ptr->an*table[ptr->n];
        c=c%m;
        ptr=ptr->next;
    }
    delete [] table;
    return c;
}

Poly operator*(const Poly& a,const Poly& b)
{
    int i,d,dega,degb,deg;
    BOOL squaring;
    ZZn t;
    Poly prod;
    term *iptr;
    term *ptr=b.start;

    squaring=FALSE;
    if (&a==&b) squaring=TRUE;

    dega=degree(a);
    deg=dega;
    if (!squaring)
    {
        degb=degree(b);
        if (degb<dega) deg=degb;
    }
    else degb=dega;
    if (deg>=FFT_BREAK_EVEN)      /* deg is minimum - both must be less than FFT_BREAK_EVEN */
    { // use fast methods 
        big *A,*B,*C;
        deg=dega+degb;     // degree of product
   
        A=(big *)mr_alloc(dega+1,sizeof(big));
        if (!squaring) B=(big *)mr_alloc(degb+1,sizeof(big));        
        C=(big *)mr_alloc(deg+1,sizeof(big));
        for (i=0;i<=deg;i++) C[i]=mirvar(0);
        ptr=a.start;
        while (ptr!=NULL)
        {
            A[ptr->n]=getbig(ptr->an);
            ptr=ptr->next;
        }

        if (!squaring)
        {
            ptr=b.start;
            while (ptr!=NULL)
            {
                B[ptr->n]=getbig(ptr->an);
                ptr=ptr->next;
            }
            mr_poly_mul(dega,A,degb,B,C);
        }
        else mr_poly_sqr(dega,A,C);
        for (d=0;d<=deg;d++)
        {
            t=C[d];
            mr_free(C[d]);
            if (t.iszero()) continue;
            prod.addterm(t,d);
        }

        mr_free(C);
        mr_free(A);
        if (!squaring) mr_free(B);

        return prod;
    }


    if (squaring)
    { // squaring
        while (ptr!=NULL)
        { // diagonal terms
            prod.addterm(ptr->an*ptr->an,ptr->n+ptr->n);
            ptr=ptr->next;
        }
        ptr=b.start;
        while (ptr!=NULL)
        { // above the diagonal
            iptr=ptr->next;
            while (iptr!=NULL)
            {
                ZZn t=ptr->an*iptr->an;
                prod.addterm(t+t,ptr->n+iptr->n);
                iptr=iptr->next;
            }
            ptr=ptr->next; 
        }
    }
    else while (ptr!=NULL)
    {
        Poly t=a; 
        t.multerm(ptr->an,ptr->n);
        ptr=ptr->next;
        prod+=t;
    }

    return prod;
}

Poly operator%(const Poly& u,const Poly&v)
{
    Poly r=u;
    ZZn m;
    term *rptr=r.start;
    term *vptr=v.start;
    term *ptr;
    if (degree(r)<degree(v)) return r;
    m=(ZZn)1/vptr->an;
    while (rptr!=NULL && rptr->n>=vptr->n)
    {
        Poly t=v;
        ZZn pq=rptr->an*m;
        int power=rptr->n-vptr->n;
        t.multerm(-pq,power);
        ptr=t.start;
        while (ptr!=NULL)
        {
            r.addterm(ptr->an,ptr->n);
            ptr=ptr->next;
        } 
        rptr=r.start;
    }
    return r;
}

Poly operator/(const Poly& u,const Poly&v)
{
    Poly q,r=u;
    term *rptr=r.start;
    term *vptr=v.start;
    term *ptr;
    while (rptr!=NULL && rptr->n>=vptr->n)
    {
        Poly t=v;
        ZZn pq=rptr->an/vptr->an;
        int power=rptr->n-vptr->n;
  // quotient
        q.addterm(pq,power);
        t.multerm(-pq,power);
        ptr=t.start;
        while (ptr!=NULL)
        {
            r.addterm(ptr->an,ptr->n);
            ptr=ptr->next;
        } 
        rptr=r.start;
    }
    return q;
}

Poly gcd(const Poly& f,const Poly& g)
{
    Poly a,b;
    a=f; b=g;
    term *ptr;
    while (b.start!=NULL)
    {
        Poly t=a%b;
        a=b;
        b=t;
    }
    ptr=a.start;
    a.multerm((ZZn)1/ptr->an,0);
    return a;
}

Poly pow(const Poly& f,int k)
{
    Poly u;
    int w,e,b;

    if (k==0)
    {
        u.addterm((ZZn)1,0);
        return u;
    }
    u=f;
    if (k==1) return u;

    e=k;
    b=0; while (k>1) {k>>=1; b++; }
    w=(1<<b);
    e-=w; w/=2;
    while (w>0)
    {
        u=(u*u);
        if (e>=w)
        {
           e-=w;
           u=(u*f);
        }
        w/=2; 
    }
    return u;
}

Poly pow(const Poly& f,const Big& k,const Poly& m)
{
    Poly u,t,u2,table[16];
    Big w,e;
    int i,j,nb,n,nbw,nzs;
    if (k==0) 
    {
        u.addterm((ZZn)1,0);
        return u;
    }
    u=(f%m);
    if (k==1) return u;

    e=k;
    w=pow((Big)2,bits(e)-1);
    e-=w; w/=2;
    while (w>0)
    {
        u=(u*u)%m;
        if (e>=w)
        {
           e-=w;
           u=(u*f)%m;
        }
        w/=2; 
    }
    return u;
}

int degree(const Poly& p)
{
    if (p.start==NULL) return 0;
    else return p.start->n;
}


BOOL iszero(Poly& p)
{
    if (degree(p)==0 && p.coeff(0)==0) return TRUE;
    else return FALSE;
}

BOOL isone(Poly& p)
{
    if (degree(p)==0 && p.coeff(0)==1) return TRUE;
    else return FALSE;
}

Poly factor(const Poly& f,int d)
{
    Poly c,u,h,g=f;
    Big r,p=get_modulus();
    while (degree(g) > d)
    {

// random monic polynomial
       u.clear();
       u.addterm((ZZn)1,2*d-1);
       for (int i=2*d-2;i>=0;i--)
       {
          r=rand(p);
          u.addterm((ZZn)r,i);
       }
       r=(pow(p,d)-1)/2;
       c=pow(u,r,g);
       c.addterm((ZZn)(-1),0);

       h=gcd(c,g);
       if (degree(h)==0 || degree(h)==degree(g)) continue;
       if (2*degree(h)>degree(g))
            g=g/h;
       else g=h;
    }
    return g;
}

void Poly::clear()
{
    term *ptr;
    while (start!=NULL)
    {   
       ptr=start->next;
       delete start;
       start=ptr;
    }
    
}

Poly& Poly::operator=(int m)
{
    clear();
    if (m!=0) addterm((ZZn)m,0);
    return *this;
}

Poly &Poly::operator=(const Poly& p)
{
    term *ptr;
    clear();
    ptr=p.start;
    while (ptr!=NULL)
    {  
        addterm(ptr->an,ptr->n);
        ptr=ptr->next;
    }    
    return *this;
    
}


Poly operator+(const Poly& a,const Poly& b)
{
    Poly sum;
    sum=a;
    sum+=b;
    return sum;
}

Poly operator-(const Poly& a,const Poly& b)
{
    Poly sum;
    sum=a;
    sum-=b;
    return sum;
}

Poly& Poly::operator+=(const Poly& p)
{
    term *ptr;
    ptr=p.start;
    while (ptr!=NULL)
    {  
        addterm(ptr->an,ptr->n);
        ptr=ptr->next;
    }    
    return *this;
}



Poly& Poly::operator*=(const ZZn& x)
{
    term *ptr=start;
    while (ptr!=NULL)
    {
        ptr->an*=x;
        ptr=ptr->next;
    }
    return *this;
}

Poly operator*(const ZZn& z,const Poly &p)
{
    Poly r=p;
    r*=z;
    return r;
}

Poly operator*(const Poly &p,const ZZn& z)
{
    Poly r=p;
    r*=z;
    return r;
}

Poly operator/(const Poly& a,const ZZn& b)
{
    Poly quo;
    quo=a;
    quo/=(ZZn)b;
    return quo;
}


Poly& Poly::operator/=(const ZZn& x)
{
    ZZn t=(ZZn)1/x;
    term *ptr=start;
    while (ptr!=NULL)
    {
        ptr->an*=t;
        ptr=ptr->next;
    }
    return *this;
}


Poly& Poly::operator-=(const Poly& p)
{
    term *ptr;
    ptr=p.start;
    while (ptr!=NULL)
    {  
        addterm(-(ptr->an),ptr->n);
        ptr=ptr->next;
    }    
    return *this;
}
 
void Poly::multerm(const ZZn& a,int power)
{
    term *ptr=start;
    while (ptr!=NULL)
    {
        ptr->an*=a;
        ptr->n+=power;
        ptr=ptr->next;
    }
}

Poly invmodxn(Poly& a,int n)
{ // Newton's method to find 1/a mod x^n
    int i,k;
    Poly b;
    k=0; while ((1<<k)<n) k++;
    b.addterm((ZZn)1/a.coeff(0),0); // important that a0 != 0
    for (i=1;i<=k;i++) b=modxn (2*b-a*b*b,1<<i);
    b=modxn(b,n);
    return b;
}

Poly modxn(const Poly& a,int n)
{ // reduce polynomial mod x^n
    Poly b;
    term* ptr=a.start;
    while (ptr!=NULL && ptr->n>=n) ptr=ptr->next;
    while (ptr!=NULL)
    {
        b.addterm(ptr->an,ptr->n);
        ptr=ptr->next;
    }
    return b;
}

Poly divxn(const Poly& a,int n)
{ // divide polynomial by x^n
    Poly b;
    term *ptr=a.start;
    while (ptr!=NULL)
    {
        if (ptr->n>=n)
            b.addterm(ptr->an,ptr->n-n);
        else break;
        ptr=ptr->next;
    }
    return b;
}

Poly reverse(const Poly& a)
{
    term *ptr=a.start;
    int deg=degree(a);
    Poly b;
    while (ptr!=NULL)
    {
        b.addterm(ptr->an,deg-ptr->n);
        ptr=ptr->next;
    } 
    return b;
}


void Poly::addterm(ZZn& a,int power)
{
    term* newone;  
    term* ptr;
    term *t,*iptr;
    ptr=start;
    iptr=NULL;
    if (a.iszero()) return;
// quick scan through to detect if term exists already
// and to find insertion point
    while (ptr!=NULL) 
    { 
        if (ptr->n==power)
        {
            ptr->an+=a;
            if (ptr->an.iszero()) 
            { // delete term
                if (ptr==start)
                { // delete first one
                    start=ptr->next;
                    delete ptr;
                    return;
                }
                iptr=ptr;
                ptr=start;
                while (ptr->next!=iptr)ptr=ptr->next;
                ptr->next=iptr->next;
                delete iptr;
                return;
            }
            return;
        }
        if (ptr->n>power) iptr=ptr;
        else break;
        ptr=ptr->next;
    }
    newone=new term;
    newone->next=NULL;
    newone->an=a;
    newone->n=power;
    if (start==NULL)
    {
        start=newone;
        return;
    }

// insert at the start

    if (iptr==NULL)
    { 
        t=start;
        start=newone;
        newone->next=t;
        return;
    }

// insert new term

    t=iptr->next;
    iptr->next=newone;
    newone->next=t;
    return;    
}

ostream& operator<<(ostream& s,const Poly& p)
{
    BOOL first=TRUE;
    ZZn a;
    term *ptr=p.start;
    if (ptr==NULL)
    { 
        s << "0";
        return s;
    }
    while (ptr!=NULL)
    {
        a=ptr->an;
        if ((Big)a<get_modulus()/2)
        {
            if (!first) s << " + ";
        }
        else 
        {
           a=(-a);
           s << " - ";
        }
        if (ptr->n==0) 
           cout << a; 
        else 
        {
            if (a!=(ZZn)1)  s << a << "*x"; 
            else            s << "x";
            if (ptr->n!=1)  s << "^" << ptr->n;
        }
        first=FALSE;
        ptr=ptr->next;
    }
    return s;
} 

