/*
 * C++ class to implement a polynomial type and to allow 
 * arithmetic on polynomials whose elements are from
 * the finite field mod p. 
 *
 * WARNING: This class has been cobbled together for a specific use with
 * the MIRACL library. It is not complete, and may not work in other 
 * applications
 *
 * This type is automatically reduced
 * wrt a polynomial modulus.
 *
 * See Knuth The Art of Computer Programming Vol.2, Chapter 4.6 
 */

#include <iostream.h>
#include "polymod.h"

Poly modulus;
  
BOOL iszero(PolyMod& m)     {return iszero(m.p);}
BOOL isone(PolyMod& m)      {return isone(m.p);}
int degree(const PolyMod& m)  {return degree(m.p);}

static int window(Big x,int i,int *nbs,int * nzs)
{ /* returns sliding window value, max. of 5 bits,         *
   * starting at i-th bit of big x. nbs is number of bits  *
   * processed, nzs is the number of additional trailing   *
   * zeros detected. Returns valid bit pattern 1x..x1 with *
   * no two adjacent 0's. So 10101 will return 21 with     *
   * nbs=5, nzs=0. 11001 will return 3, with nbs=2, nzs=2, *
   * having stopped after the first 11..  */

    int j,r,w;
    w=5;

/* check for leading 0 bit */

    *nbs=1;
    *nzs=0;

    if (!bit(x,i)) return 0;

/* adjust window size if not enough bits left */
   
    if (i-w+1<0) w=i+1;

    r=1;
    for (j=i-1;j>i-w;j--)
    { /* accumulate bits. Abort if two 0's in a row */
        (*nbs)++;
        r*=2;
        if (bit(x,j)) r+=1;
        if (r%4==0)
        { /* oops - too many zeros - shorten window */
            r/=4;
            *nbs-=2;
            *nzs=2;
            break;
        }
    }
    if (r%2==0)
    { /* remove trailing 0 */
        r/=2;
        *nzs=1;
        (*nbs)--;
    }
    return r;
}

PolyMod reduce(const Poly& p)
{
    int i,d;
    ZZn t;
    PolyMod rem;
    big *G,*R;
    term *ptr;
    int n=degree(p);
    int degm=degree(modulus);
    if (n-degm < FFT_BREAK_EVEN) return (PolyMod)p;

    G=(big *)mr_alloc(n+1,sizeof(big));
    R=(big *)mr_alloc(degm,sizeof(big));
    for (i=0;i<degm;i++) R[i]=mirvar(0);

    ptr=p.start;
    while (ptr!=NULL)
    {
        G[ptr->n]=getbig(ptr->an);
        ptr=ptr->next; 
    }
    if (!mr_poly_rem(n,G,R))
    {  // reset the modulus - things have changed
        setmod(modulus);
        mr_poly_rem(n,G,R);
    }
 
    for (d=0;d<degm;d++)
    {
        t=R[d];
        mr_free(R[d]);
        if (t.iszero()) continue;
        rem.addterm(t,d);
    }

    mr_free(R);
    mr_free(G);

    return rem;
}

void setmod(const Poly& p) 
{ 
    int n;
    Poly h;
    term *ptr;
    big *f,*rf;
    modulus=p;
    n=degree(p);
    if (n<FFT_BREAK_EVEN) return;
    
    h=reverse(p);
    h=invmodxn(h,n);
    h=reverse(h);   // h=RECIP(f)

    f=(big *)mr_alloc(n+1,sizeof(big));
    rf=(big *)mr_alloc(n+1,sizeof(big));
 
    ptr=p.start;
    while (ptr!=NULL)
    {
       f[ptr->n]=getbig(ptr->an);
       ptr=ptr->next;
    }   
    ptr=h.start;
    while (ptr!=NULL)
    {
       rf[ptr->n]=getbig(ptr->an);
       ptr=ptr->next;
    }   
 
    mr_polymod_set(n,rf,f);

    mr_free(rf);
    mr_free(f);
}

PolyMod operator*(const PolyMod& a,const PolyMod& b) 
                                     {return reduce(a.p*b.p);}

PolyMod operator-(const PolyMod& a,const PolyMod& b)
                                     {return (a.p-b.p);}
PolyMod operator+(const PolyMod& a,const PolyMod& b)
                                     {return (a.p+b.p);}
PolyMod operator*(const PolyMod& a,const ZZn& z)
                                     {return (z*a.p);}
PolyMod operator*(const ZZn& z,const PolyMod& m)
                                     {return (z*m.p);}

    
PolyMod operator/(const PolyMod& a,const ZZn& z)
                                     {return (a.p/z);}

Poly gcd(const PolyMod& m)
                                     {return gcd(m.p,modulus);}  

PolyMod pow(const PolyMod& f,const Big& k)
{
    PolyMod u,u2,table[16];
    Big w,e;
    int i,j,nb,n,nbw,nzs;    
    if (k==0) 
    {
        u.addterm((ZZn)1,0);
        return u;
    }
    u=reduce(f.p);
    if (k==1) return u;

    if (degree(f)>=FFT_BREAK_EVEN)
    {
        u2=(u*u);

        table[0]=u;
        for (i=1;i<16;i++)
            table[i]=u2*table[i-1];
        nb=bits(k);
        if (nb>1) for (i=nb-2;i>=0;)
        {
            n=window(k,i,&nbw,&nzs);
            for (j=0;j<nbw;j++)
            u=(u*u);

            if (n>0) u=u*table[n/2];
            i-=nbw;
            if (nzs)
            {
                for (j=0;j<nzs;j++) u=u*u;
                i-=nzs;
            }        
        }
    }
    else
    {
        e=k;
        w=pow((Big)2,bits(e)-1);
        e-=w; w/=2;
        while (w>0)
        {
            u=u*u;
            if (e>=w)
            {
               e-=w;
               u=u*f;
            }
            w/=2; 
        }
    }
    return u;
}   

ostream& operator<<(ostream& s,const PolyMod& m) 
                                     { s << m.p; return s;} 

