/*
 * C++ class to implement a bivariate polynomial type and to allow 
 * arithmetic on such polynomials whose coefficients are from
 * the finite field mod p
 *
 * WARNING: This class has been cobbled together for a specific use with
 * the MIRACL library. It is not complete, and may not work in other 
 * applications
 *
 * See Knuth The Art of Computer Programming Vol.2, Chapter 4.6 
 */

#include "polyxy.h"

PolyXY::PolyXY(const PolyXY& p)
{
    termXY *ptr=p.start;
    start=NULL;
    while (ptr!=NULL)
    {  
        addterm(ptr->an,ptr->nx,ptr->ny);
        ptr=ptr->next;
    }    
}

PolyXY::~PolyXY()
{
   termXY *nx;
   while (start!=NULL)
   {   
       nx=start->next;
       delete start;
       start=nx;
   }
}

ZZn PolyXY::coeff(int powx,int powy)
{
    ZZn c=0;
    termXY *ptr=start;
    while (ptr!=NULL)
    {
        if (ptr->nx==powx && ptr->ny==powy)
        {
            c=ptr->an;
            return c;
        }
        ptr=ptr->next;
    }
    return c;
}

PolyXY diff_dx(const PolyXY& f)
{
    PolyXY r;
    termXY *ptr=f.start;
    while (ptr!=NULL)
    {
        r.addterm(ptr->an*ptr->nx,ptr->nx-1,ptr->ny);
        ptr=ptr->next;
    }
    return r;
}

PolyXY diff_dy(const PolyXY& f)
{
    PolyXY r;
    termXY *ptr=f.start;
    while (ptr!=NULL)
    {
        r.addterm(ptr->an*ptr->ny,ptr->nx,ptr->ny-1);
        ptr=ptr->next;
    }
    return r;
}

ZZn PolyXY::F(const ZZn& x,const ZZn& y)
{
    ZZn f=0;
    termXY *ptr=start;
    while (ptr!=NULL)
    {
        f+=(ptr->an*pow(x,ptr->nx)*pow(y,ptr->ny));
        ptr=ptr->next;
    }
    return f;
}

Poly PolyXY::F(const ZZn& y)
{
    Poly r;
    termXY *ptr=start;
    while (ptr!=NULL)
    {
        r.addterm(ptr->an*pow(y,ptr->ny),ptr->nx);
        ptr=ptr->next;
    }
    return r;
}

void PolyXY::clear()
{
    termXY *ptr;
    while (start!=NULL)
    {   
       ptr=start->next;
       delete start;
       start=ptr;
    }
    
}

PolyXY& PolyXY::operator=(const PolyXY& p)
{
    termXY *ptr;
    clear();
    ptr=p.start;
    while (ptr!=NULL)
    {  
        addterm(ptr->an,ptr->nx,ptr->ny);
        ptr=ptr->next;
    }    
    return *this;
    
}

void PolyXY::addterm(ZZn& a,int powx,int powy)
{
    termXY* newone;  
    termXY* ptr;
    termXY *t,*iptr;
    ptr=start;
    iptr=NULL;
    if (a.iszero()) return;

// quick scan through to detect if term exists already
// and to find insertion point
    while (ptr!=NULL) 
    { 
        if (ptr->nx==powx && ptr->ny==powy)
        {
            ptr->an+=a;
            if (ptr->an.iszero()) 
            { // delete term
                if (ptr==start)
                { // delete first one
                    start=ptr->next;
                    delete ptr;
                    return;
                }
                iptr=ptr;
                ptr=start;
                while (ptr->next!=iptr)ptr=ptr->next;
                ptr->next=iptr->next;
                delete iptr;
                return;
            }
            return;
        }
        if (ptr->nx>powx) iptr=ptr;
        if (ptr->nx==powx && ptr->ny>powy) iptr=ptr;
        ptr=ptr->next;
    }
    newone=new termXY;
    newone->next=NULL;
    newone->an=a;
    newone->nx=powx;
    newone->ny=powy;

    if (start==NULL)
    {
        start=newone;
        return;
    }

// insert at the start

    if (iptr==NULL)
    { 
        t=start;
        start=newone;
        newone->next=t;
        return;
    }

// insert new term

    t=iptr->next;
    iptr->next=newone;
    newone->next=t;
    return;    
}

ostream& operator<<(ostream& s,const PolyXY& p)
{
    BOOL first=TRUE;
    ZZn a;
    termXY *ptr=p.start;
    if (ptr==NULL)
    { 
        s << "0";
        return s;
    }
    while (ptr!=NULL)
    {
        a=ptr->an;
        if ((Big)a<get_modulus()/2)
        {
            if (!first) s << " + ";
        }
        else 
        {
           a=(-a);
           s << " - ";
        }
        if (ptr->nx==0 && ptr->ny==0) 
           cout << a; 
        else 
        {
            if (a!=(ZZn)1) s << a << "*";
            if (ptr->nx!=0)
            {
                s << "x";
                if (ptr->nx!=1)  s << "^" << ptr->nx;
            }
            if (ptr->ny!=0)
            {
                if (ptr->nx!=0) s << ".";
                s << "y";
                if (ptr->ny!=1)  s << "^" << ptr->ny;
            }
        }
        first=FALSE;
        ptr=ptr->next;
    }
    return s;
} 

