/*
 *   Proposed Digital Signature Standard
 *
 *   Elliptic Curve Variation - See Dr. Dobbs Journal April 1997
 *
 *   This program generates one set of public and private keys in files 
 *   public.ecs and private.ecs respectively. Notice that the public key 
 *   can be much shorter in this scheme, for the same security level.
 *
 *   It is assumed that Curve parameters are to be found in file common.ecs
 *
 *   For the curve y^2=x^3+Ax+b mod p, these parameters are:
 *
 *   A,B,x,y,p,q where and b are parameters of the equation above, (x,y)
 *   is an initial point on the curve, p is a prime, and q is the order of 
 *   the (x,y) point, itself a large prime
 * 
 *   Copyright (c) 1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

int main()
{
    FILE *fp;
    int ep;
    epoint *g;
    big a,b,p,q,x,y,d;
    long seed;
    mirsys(50,0);
    a=mirvar(0);
    b=mirvar(0);
    p=mirvar(0);
    q=mirvar(0);
    x=mirvar(0);
    y=mirvar(0);
    d=mirvar(0);
/* randomise */
    printf("Enter 9 digit random number seed  = ");
    scanf("%ld",&seed);
    getchar();
    irand(seed);
    fp=fopen("common.ecs","r");
    if (fp==NULL)
    {
        printf("file common.ecs does not exist\n");
        return 0;
    }
    cinnum(a,fp);
    cinnum(b,fp);
    cinnum(x,fp);
    cinnum(y,fp);
    cinnum(p,fp);
    cinnum(q,fp);
    fclose(fp);
    ecurve_init(a,b,p,MR_PROJECTIVE);  /* initialise curve */
    g=epoint_init();
    epoint_set(x,y,0,g); /* initialise point of order q */

/* generate public/private keys */

    bigrand(q,d);
    ecurve_mult(d,g,g);
    
    ep=epoint_get(g,x,x); /* compress point */

    printf("public key = %d ",ep);
    cotnum(x,stdout);

    fp=fopen("public.ecs","w");
    fprintf(fp,"%d ",ep);
    cotnum(x,fp);
    fclose(fp);

    fp=fopen("private.ecs","w");
    cotnum(d,fp);
    fclose(fp);
    return 0;
}

