/*
 *   Proposed Digital Signature Standard
 *
 *   Elliptic Curve Variation - See Dr. Dobbs Journal April 1997
 *
 *   This program generates one set of public and private keys in files 
 *   public.ecs and private.ecs respectively. Notice that the public key 
 *   can be much shorter in this scheme, for the same security level.
 *
 *   It is assumed that Curve parameters are to be found in file common.ecs
 *
 *   For the curve y^2=x^3+Ax+b mod p, these parameters are:
 *
 *   A,B,x,y,p,q where A and B are parameters of the equation above, (x,y)
 *   is an initial point on the curve, p is a prime, and q is the order of 
 *   the (x,y) point, itself a large prime.
 * 
 *   Copyright (c) 1997 Shamus Software Ltd.
 */

#include <iostream.h>
#include <fstream.h>
#include <elliptic.h>

Miracl precision=20;

int main()
{
    ifstream common("common.ecs");    /* construct file I/O streams */
    ofstream public_key("public.ecs");
    ofstream private_key("private.ecs");
    int ep;
    ECn G;
    Big a,b,p,q,x,y,d;
    long seed;
    cout << "Enter 9 digit random number seed  = ";
    cin >> seed;
    irand(seed);
    common >> a >> b >> x >> y >> p >> q;
    ecurve(a,b,p,MR_PROJECTIVE);
    G=ECn(x,y);

/* generate public/private keys */

    d=rand(q);
    G*=d;
    ep=G.get(x);
    cout << "public key = " << ep << " " << x << endl;
    public_key << ep << " " << x << endl;
    private_key << d << endl;
    return 0;
}

