/*
 *   Proposed Digital Signature Standard (DSS)
 *
 *   Elliptic Curve Variation - See Dr. Dobbs Journal April 1997
 *
 *   This program verifies the signature given to a <file> in
 *   <file>.ecs generated by program ecsign
 * 
 *   Copyright (c) 1997 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>
#include <stdlib.h>
#include <string.h>

void strip(char *name)
{ /* strip off filename extension */
    int i;
    for (i=0;name[i]!='\0';i++)
    {
        if (name[i]!='.') continue;
        name[i]='\0';
        break;
    }
}

static void hashing(FILE *fp,big hash)
{ /* compute hash function */
    char h[20];
    sha sh;
    int i,ch;
    shs_init(&sh);
    while ((ch=fgetc(fp))!=EOF) shs_process(&sh,ch);
    shs_hash(&sh,h);
/* convert to big number */
    get_mip()->INPLEN=20;
    instr(hash,h);
}

int main()
{
    FILE *fp;
    int ep;
    epoint *g,*public;
    char ifname[13],ofname[13];
    big a,b,p,q,x,y,v,u1,u2,r,s,hash;
    mirsys(200,256);
    a=mirvar(0);
    b=mirvar(0);
    p=mirvar(0);
    q=mirvar(0);
    x=mirvar(0);
    y=mirvar(0);
    v=mirvar(0);
    u1=mirvar(0);
    u2=mirvar(0);
    s=mirvar(0);
    r=mirvar(0);
    hash=mirvar(0);
/* get public data */
    fp=fopen("common.ecs","r");
    if (fp==NULL)
    {
        printf("file common.ecs does not exist\n");
        return 0;
    }
    cinnum(a,fp);
    cinnum(b,fp);
    cinnum(x,fp);
    cinnum(y,fp);
    cinnum(p,fp);
    cinnum(q,fp);

    fclose(fp);

    ecurve_init(a,b,p,MR_PROJECTIVE);  /* initialise curve */
    g=epoint_init();
    epoint_set(x,y,0,g); /* initialise point of order q */

/* get public key of signer */
    fp=fopen("public.ecs","r");
    if (fp==NULL)
    {
        printf("file public.ecs does not exist\n");
        return 0;
    }
    fscanf(fp,"%d",&ep);
    cinnum(x,fp);
    fclose(fp);

    public=epoint_init();
    epoint_set(x,x,ep,public);  /* decompress */

/* get message */
    printf("signed file = ");
    gets(ifname);
    strcpy(ofname,ifname);
    strip(ofname);
    strcat(ofname,".ecs");
    if ((fp=fopen(ifname,"rb"))==NULL)
    { /* no message */
        printf("Unable to open file %s\n",ifname);
        return 0;
    }
    hashing(fp,hash);
    fclose(fp);
    fp=fopen(ofname,"r");
    if (fp==NULL)
    { /* no signature */
        printf("signature file %s does not exist\n",ofname);
        return 0;
    }
    cinnum(r,fp);
    cinnum(s,fp);
    fclose(fp);
    if (compare(r,q)>=0 || compare(s,q)>=0)
    {
        printf("Signature is NOT verified\n");
        return 0;
    }
    xgcd(s,q,s,s,s);
    mad(hash,s,s,q,q,u1);
    mad(r,s,s,q,q,u2);

    ecurve_mult2(u2,public,u1,g,g);
    epoint_get(g,v,v);
    divide(v,q,q);
    if (compare(v,r)==0) printf("Signature is verified\n");
    else                 printf("Signature is NOT verified\n");
    return 0;
}

